﻿using System;
using System.Collections.Generic;
using HeuristicLab.Encodings.RealVectorEncoding;

namespace HeuristicLab.Problems.MultiObjectiveTestFunctions {

  /// <summary>
  /// Crowding distance d(x,A) is usually defined between a point x and a set of points  A
  /// d(x,A) is then a weighted sum over all dimensions where for each dimension the next larger and the next smaller Point to x are subtracted
  /// I extended the concept and defined the Crowding distance of a front A as the mean of the crowding distances of every point x in A 
  /// C(A) = mean(x,A) where x in A
  /// </summary>
  public class Crowding : IMultiObjectiveDistance {
    private double[][] bounds;
    public Crowding(double[][] bounds) {
      this.bounds = bounds;
    }


    public static double GetDistance(RealVector[] front, RealVector[] optimalFront, double[][] bounds) {
      return new Crowding(bounds).Compare(front, optimalFront);
    }


    public static double GetCrowding(RealVector[] front, double[][] bounds) {
      return new Crowding(bounds).Get(front);
    }


    public double Get(RealVector[] front) {
      double sum = 0;
      foreach(RealVector point in front) {
        sum += CalcCrowding(point, front);
      }
      return sum / front.Length;
    }

    public double Compare(RealVector[] front, RealVector[] optimalFront) {
      return Get(optimalFront) - Get(front);
    }

    private double CalcCrowding(RealVector point, RealVector[] list) {
      double sum = 0;
      for (int i = 0; i < point.Length; i++) {
        sum += CalcCrowding(point, list, i);
      }
      return sum;
    }

    private double CalcCrowding(RealVector point, RealVector[] list, int dim) {
      double fmax = bounds[dim % bounds.Length][1];
      double fmin = bounds[dim % bounds.Length][0];

      Array.Sort<RealVector>(list, new DimensionComparer(dim, false));
      if (list[0][dim] == point[0] || list[list.Length - 1][dim] == point[0]) return Double.PositiveInfinity;
      int pos = binarySearch(point[dim], list, dim);
      return (list[pos + 1][dim] - list[pos - 1][dim] )/ (fmax - fmin);
    }

    private int binarySearch(double x, RealVector[] list, int dim) {
      int low = 0;
      int high = list.Length - 1;
      while (high >= low) {
        int middle = (low + high) / 2;
        if (list[middle][dim] == x) return middle;
        if (list[middle][dim] < x)  low = middle + 1;
        if (list[middle][dim] > x)  high = middle - 1; 
      }
      return -1; //This should never happen
    }

    private class DimensionComparer : IComparer<RealVector> {
      private int dim;
      private int descending;

      public DimensionComparer(int dimension, bool descending) {
        this.dim = dimension;
        this.descending = descending ? -1 : 1;
      }

      #region IComparer<DoubleArray> Members

      public int Compare(RealVector x, RealVector y) {
        if (x[dim] < y[dim]) return -descending;
        else if (x[dim] > y[dim]) return descending;
        else return 0;
      }

      #endregion
    }

  }
}
