﻿using System;
using System.Collections.Generic;
using HeuristicLab.Encodings.RealVectorEncoding;

namespace HeuristicLab.Problems.MultiObjectiveTestFunctions {
  /// <summary>
  /// The Hyprevolume-metric is defined as the Hypervolume enclosed between a given reference point,
  /// that is fixed for every evaluation function and the evaluated front.
  /// 
  /// Example:
  /// r is the reference Point at (1|1)  and every Point p is part of the evaluated front
  /// The filled Area labled HV is the 2 diensional Hypervolume enclosed by this front. 
  /// 
  /// (0|1)                (1|1)
  ///   +      +-------------r
  ///   |      |###### HV ###|
  ///   |      p------+######|
  ///   |             p+#####|
  ///   |              |#####|
  ///   |              p-+###|
  ///   |                p---+
  ///   |                 
  ///   +--------------------1
  /// (0|0)                (1|0)
  /// 
  ///  Please note that in this example both dimensions are minimized. The reference Point need to be dominated by EVERY point in the evaluated front 
  /// 
  /// </summary>
  public class Hypervolume : IMultiObjectiveDistance {

    private RealVector reference;
    private bool[] maximization;
    public Hypervolume(RealVector reference, bool[] maximization) {
      if (reference.Length != 2) throw new Exception("Only 2-dimensional cases are supported yet");
      this.reference = reference;
      this.maximization = maximization;
    }

    public double Compare(RealVector[] front, RealVector[] optimalFront) {
      return GetHypervolume(optimalFront) - GetHypervolume(front);

    }

    public double GetHypervolume(RealVector[] front) {
      //TODO what to do if set contains dominated points
      front = (RealVector[])front.Clone(); //TODO this seems shady
      Array.Sort<RealVector>(front, new DimensionComparer(0, maximization[0]));
      RealVector last = front[front.Length - 1];
      CheckConsistency(last, 0);
      CheckConsistency(last, 1);
      double sum = 0;
      for (int i = 0; i < front.Length - 1; i++) {
        CheckConsistency(front[i], 1);
        sum += Math.Abs((front[i][0] - front[i + 1][0])) * Math.Abs((front[i][1] - reference[1]));
      }

      sum += Math.Abs(reference[0] - last[0]) * Math.Abs(reference[1] - last[1]);
      return sum;
    }

    public static double GetHypervolume(RealVector[] front, RealVector reference, bool[] maximization){
      Hypervolume comp = new Hypervolume(reference, maximization);
      return comp.GetHypervolume(front);
    }

    public static double GetDistance(RealVector[] front, RealVector[] optimalFront, RealVector reference, bool[] maximization) {
      return GetHypervolume(optimalFront, reference, maximization) - GetHypervolume(front, reference, maximization);
    }

    private void CheckConsistency(RealVector point, int dim) {
      if (!maximization[dim] && point[dim] > reference[dim]) throw new Exception("Reference Point must be dominated by all points of the front");
      if (maximization[dim] && point[dim] < reference[dim]) throw new Exception("Reference Point must be dominated by all points of the front");
      if (point.Length != 2) throw new Exception("Only 2-dimensional cases are supported yet");
    }

    private class DimensionComparer : IComparer<RealVector> {
      private int dim;
      private int descending;

      public DimensionComparer(int dimension, bool descending) {
        this.dim = dimension;
        this.descending = descending ? -1 : 1;
      }

      #region IComparer<DoubleArray> Members

      public int Compare(RealVector x, RealVector y) {
        if (x[dim] < y[dim]) return -descending;
        else if (x[dim] > y[dim]) return descending;
        else return 0;
      }

      #endregion
    }

  }
}
