﻿using System;
using System.Collections.Generic;
using HeuristicLab.Encodings.RealVectorEncoding;

namespace HeuristicLab.Problems.MultiObjectiveTestFunctions {
  /// <summary>
  /// The Hyprevolume-metric is defined as the Hypervolume enclosed between a given reference point,
  /// that is fixed for every evaluation function and the evaluated front.
  /// 
  /// Example:
  /// r is the reference Point at (1|1)  and every Point p is part of the evaluated front
  /// The filled Area labled HV is the 2 diensional Hypervolume enclosed by this front. 
  /// 
  /// (0|1)                (1|1)
  ///   +      +-------------r
  ///   |      |###### HV ###|
  ///   |      p------+######|
  ///   |             p+#####|
  ///   |              |#####|
  ///   |              p-+###|
  ///   |                p---+
  ///   |                 
  ///   +--------------------1
  /// (0|0)                (1|0)
  /// 
  ///  Please note that in this example both dimensions are minimized. The reference Point need to be dominated by EVERY point in the evaluated front 
  /// 
  /// </summary>
  public class FastHypervolume {

    private RealVector reference;
    private bool[] maximization;
    public FastHypervolume(RealVector reference, bool[] maximization) {
      if (reference.Length != 2) throw new Exception("Only 2-dimensional cases are supported yet");
      this.reference = reference;
      this.maximization = maximization;
    }

    public double Compare(RealVector[] front, RealVector[] optimalFront, double[][] bounds) {
      return GetHypervolume(optimalFront, bounds) - GetHypervolume(front, bounds);

    }


    /// <summary>
    /// 
    /// </summary>
    /// <param name="front"></param>
    /// <param name="bounds">also called region</param>
    /// <returns></returns>
    public double GetHypervolume(RealVector[] front, double[][] bounds) {
      //TODO what to do if set contains dominated points
      var list = new List<RealVector>();
      list.AddRange(front);
      list.Sort(new DimensionComparer(reference.Length - 1, maximization[reference.Length - 1]));
      var results = new ExPrivates(this);
      GetHV(DeepClone(bounds), list, 1, reference[reference.Length - 1], results);
      return results.Volume;

    }

    private double[][] DeepClone(double[][] array) {
      double[][] copy = new double[array.Length][];
      for (int i = 0; i < copy.Length; i++) {
        copy[i] = new double[array[i].Length];
        for (int j = 0; i < array[i].Length; j++) {
          copy[i][j] = array[i][j];
        }
      }
      return copy;
    }

    private class ExPrivates {
      private double volume;
      public double Volume {
        get { return volume; }
        set { volume = value; }
      }
      private int d;

      public int D {
        get { return D; }
      }
      public ExPrivates(FastHypervolume hv) {
        volume = 0;
        d = hv.reference.Length;
      }



    }

    private void GetHV(double[][] region, List<RealVector> points, int split, double cover, ExPrivates results) {
      double coverNew = cover;
      int coverIndex = 1;
      bool allPiles = true;
      int bound = -1;

      /* is the region completely covered? */
      while (coverNew == cover && coverIndex != points.Count) {
        if (covers(points[coverIndex], region, results)) {
          coverNew = points[coverIndex][results.D];
          results.Volume += getMeasure(region) * (cover - coverNew);
        } else { }
        coverIndex++;
      }
      if (coverIndex == 1) return;

      for (int i = 1; i < coverIndex; i++) { if (checkPile(points[i], region) == -1) allPiles = false; }

      if (allPiles) {
        /* calculate volume by sweeping along dimension d */
        var trellis = new double[reference.Length];
        int i = 1;
        for (int j = 1; j < results.D; j++) trellis[j] = reference[j];

        double next;//bernhard
        do {
          double current = points[i][results.D];
          do {
            int pile = getPile(points[i], region);
            if (points[i][pile] < trellis[pile]) trellis[pile] = points[i][pile];
            i++;
            if (i < coverIndex - 1) next = points[i][results.D]; else next = coverNew;
          } while (current == next);
          results.Volume += measure(trellis, region) * (next - current);
        } while (next != coverNew);

      } else {
        /* split region in two children regions */
        do {
          var intersect = new List<Double>(); var nonIntersect = new List<Double>();
          for (int i = 1; i < coverIndex; i++) {
            var intersection = intersects(points[i], region, split);
            if (intersection == 1) intersect.Add(points[i][split]);
            if (intersection == 0) nonIntersect.Add(points[i][split]);
          }
          if (intersect.Count != 0) bound = median(intersect);
          else if (nonIntersect.Count > Math.Sqrt(n)) bound = median(nonIntersect);
          else split++;
        } while (bound == -1);
      }


      var regionC = DeepClone(region);
      regionC[split][1] = bound;
      var pointsC = new List<RealVector>();
      for (int i = 1; i < coverIndex; i++) {
        if (partCovers(points[i], regionC, results)) move(points[i], pointsC);
      }
      if (pointsC.Count != 0) GetHV(regionC, pointsC, split, coverNew, results);
      reinsert(pointsC, points);
      regionC = region;
      regionC[split][0] = bound;
      pointsC = new List<RealVector>();
      for (int i = 1; i < coverIndex; i++) {
        if (partCovers(points[i], regionC, results)) move(points[i], pointsC);
      }
      if (pointsC.Count != 0) GetHV(regionC, pointsC, split, coverNew, results);
      reinsert(pointsC, points);
    }

    private int checkPile(RealVector point, double[][] region, ExPrivates results) {
      int pile = -1;
      for (int j = 1; j < results.D; j++) {
        if (point[j] > region[j][0]) {
          if (pile != -1) return -1;
        }
        pile = j;
      }
      return pile;

    }

    private int intersects(RealVector point, double[][] region, int split) {
      if (region[split][0] >= point[split]) return -1;
      for (int j = 1; j < split; j++) {
        if (point[j] > region[j][0]) return 1;
      }
      return 0;

    }

    private bool covers(RealVector point, double[][] region, ExPrivates results) {
      for (int j = 1; j < results.D; j++) {
        if (point[j] > region[j][0]) return false;
      }
      return true;

    }

    private bool partCovers(RealVector point, double[][] region, ExPrivates results) {
      for (int j = 1; j < results.D; j++) {
        if (point[j] >= region[j][1]) return false;
      }
      return true;
    }

    public static double GetHypervolume(RealVector[] front, RealVector reference, bool[] maximization) {
      Hypervolume comp = new Hypervolume(reference, maximization);
      return comp.GetHypervolume(front);
    }

    public static double GetDistance(RealVector[] front, RealVector[] optimalFront, RealVector reference, bool[] maximization) {
      return GetHypervolume(optimalFront, reference, maximization) - GetHypervolume(front, reference, maximization);
    }

    private void CheckConsistency(RealVector point, int dim) {
      if (!maximization[dim] && point[dim] > reference[dim]) throw new Exception("Reference Point must be dominated by all points of the front");
      if (maximization[dim] && point[dim] < reference[dim]) throw new Exception("Reference Point must be dominated by all points of the front");
    }

    private class DimensionComparer : IComparer<RealVector> {
      private int dim;
      private int descending;

      public DimensionComparer(int dimension, bool descending) {
        this.dim = dimension;
        this.descending = descending ? -1 : 1;
      }

      #region IComparer<DoubleArray> Members

      public int Compare(RealVector x, RealVector y) {
        if (x[dim] < y[dim]) return -descending;
        else if (x[dim] > y[dim]) return descending;
        else return 0;
      }

      #endregion
    }

  }
}
